EXPECTED = {'PKIXCRMF-2005': {'extensibility-implied': False,
                   'imports': {'CryptographicMessageSyntax2004': ['EnvelopedData'],
                               'PKIX1Explicit88': ['Version',
                                                   'AlgorithmIdentifier',
                                                   'Name',
                                                   'Time',
                                                   'SubjectPublicKeyInfo',
                                                   'Extensions',
                                                   'UniqueIdentifier',
                                                   'Attribute'],
                               'PKIX1Implicit88': ['GeneralName']},
                   'object-classes': {},
                   'object-sets': {},
                   'tags': 'IMPLICIT',
                   'types': {'AttributeTypeAndValue': {'members': [{'name': 'type',
                                                                    'type': 'OBJECT '
                                                                            'IDENTIFIER'},
                                                                   {'choices': {},
                                                                    'name': 'value',
                                                                    'type': 'ANY '
                                                                            'DEFINED '
                                                                            'BY',
                                                                    'value': 'type'}],
                                                       'type': 'SEQUENCE'},
                             'Attributes': {'element': {'type': 'Attribute'},
                                            'type': 'SET OF'},
                             'Authenticator': {'type': 'UTF8String'},
                             'CertId': {'members': [{'name': 'issuer',
                                                     'type': 'GeneralName'},
                                                    {'name': 'serialNumber',
                                                     'type': 'INTEGER'}],
                                        'type': 'SEQUENCE'},
                             'CertReq': {'type': 'CertRequest'},
                             'CertReqMessages': {'element': {'type': 'CertReqMsg'},
                                                 'size': [(1, 'MAX')],
                                                 'type': 'SEQUENCE OF'},
                             'CertReqMsg': {'members': [{'name': 'certReq',
                                                         'type': 'CertRequest'},
                                                        {'name': 'popo',
                                                         'optional': True,
                                                         'type': 'ProofOfPossession'},
                                                        {'element': {'type': 'AttributeTypeAndValue'},
                                                         'name': 'regInfo',
                                                         'optional': True,
                                                         'size': [(1, 'MAX')],
                                                         'type': 'SEQUENCE '
                                                                 'OF'}],
                                            'type': 'SEQUENCE'},
                             'CertRequest': {'members': [{'name': 'certReqId',
                                                          'type': 'INTEGER'},
                                                         {'name': 'certTemplate',
                                                          'type': 'CertTemplate'},
                                                         {'name': 'controls',
                                                          'optional': True,
                                                          'type': 'Controls'}],
                                             'type': 'SEQUENCE'},
                             'CertTemplate': {'members': [{'name': 'version',
                                                           'optional': True,
                                                           'tag': {'number': 0},
                                                           'type': 'Version'},
                                                          {'name': 'serialNumber',
                                                           'optional': True,
                                                           'tag': {'number': 1},
                                                           'type': 'INTEGER'},
                                                          {'name': 'signingAlg',
                                                           'optional': True,
                                                           'tag': {'number': 2},
                                                           'type': 'AlgorithmIdentifier'},
                                                          {'name': 'issuer',
                                                           'optional': True,
                                                           'tag': {'number': 3},
                                                           'type': 'Name'},
                                                          {'name': 'validity',
                                                           'optional': True,
                                                           'tag': {'number': 4},
                                                           'type': 'OptionalValidity'},
                                                          {'name': 'subject',
                                                           'optional': True,
                                                           'tag': {'number': 5},
                                                           'type': 'Name'},
                                                          {'name': 'publicKey',
                                                           'optional': True,
                                                           'tag': {'number': 6},
                                                           'type': 'SubjectPublicKeyInfo'},
                                                          {'name': 'issuerUID',
                                                           'optional': True,
                                                           'tag': {'number': 7},
                                                           'type': 'UniqueIdentifier'},
                                                          {'name': 'subjectUID',
                                                           'optional': True,
                                                           'tag': {'number': 8},
                                                           'type': 'UniqueIdentifier'},
                                                          {'name': 'extensions',
                                                           'optional': True,
                                                           'tag': {'number': 9},
                                                           'type': 'Extensions'}],
                                              'type': 'SEQUENCE'},
                             'Controls': {'element': {'type': 'AttributeTypeAndValue'},
                                          'size': [(1, 'MAX')],
                                          'type': 'SEQUENCE OF'},
                             'EncKeyWithID': {'members': [{'name': 'privateKey',
                                                           'type': 'PrivateKeyInfo'},
                                                          {'members': [{'name': 'string',
                                                                        'type': 'UTF8String'},
                                                                       {'name': 'generalName',
                                                                        'type': 'GeneralName'}],
                                                           'name': 'identifier',
                                                           'optional': True,
                                                           'type': 'CHOICE'}],
                                              'type': 'SEQUENCE'},
                             'EncryptedKey': {'members': [{'name': 'encryptedValue',
                                                           'type': 'EncryptedValue'},
                                                          {'name': 'envelopedData',
                                                           'tag': {'number': 0},
                                                           'type': 'EnvelopedData'}],
                                              'type': 'CHOICE'},
                             'EncryptedValue': {'members': [{'name': 'intendedAlg',
                                                             'optional': True,
                                                             'tag': {'number': 0},
                                                             'type': 'AlgorithmIdentifier'},
                                                            {'name': 'symmAlg',
                                                             'optional': True,
                                                             'tag': {'number': 1},
                                                             'type': 'AlgorithmIdentifier'},
                                                            {'name': 'encSymmKey',
                                                             'optional': True,
                                                             'tag': {'number': 2},
                                                             'type': 'BIT '
                                                                     'STRING'},
                                                            {'name': 'keyAlg',
                                                             'optional': True,
                                                             'tag': {'number': 3},
                                                             'type': 'AlgorithmIdentifier'},
                                                            {'name': 'valueHint',
                                                             'optional': True,
                                                             'tag': {'number': 4},
                                                             'type': 'OCTET '
                                                                     'STRING'},
                                                            {'name': 'encValue',
                                                             'type': 'BIT '
                                                                     'STRING'}],
                                                'type': 'SEQUENCE'},
                             'KeyGenParameters': {'type': 'OCTET STRING'},
                             'OldCertId': {'type': 'CertId'},
                             'OptionalValidity': {'members': [{'name': 'notBefore',
                                                               'optional': True,
                                                               'tag': {'number': 0},
                                                               'type': 'Time'},
                                                              {'name': 'notAfter',
                                                               'optional': True,
                                                               'tag': {'number': 1},
                                                               'type': 'Time'}],
                                                  'type': 'SEQUENCE'},
                             'PBMParameter': {'members': [{'name': 'salt',
                                                           'type': 'OCTET '
                                                                   'STRING'},
                                                          {'name': 'owf',
                                                           'type': 'AlgorithmIdentifier'},
                                                          {'name': 'iterationCount',
                                                           'type': 'INTEGER'},
                                                          {'name': 'mac',
                                                           'type': 'AlgorithmIdentifier'}],
                                              'type': 'SEQUENCE'},
                             'PKIArchiveOptions': {'members': [{'name': 'encryptedPrivKey',
                                                                'tag': {'number': 0},
                                                                'type': 'EncryptedKey'},
                                                               {'name': 'keyGenParameters',
                                                                'tag': {'number': 1},
                                                                'type': 'KeyGenParameters'},
                                                               {'name': 'archiveRemGenPrivKey',
                                                                'tag': {'number': 2},
                                                                'type': 'BOOLEAN'}],
                                                   'type': 'CHOICE'},
                             'PKIPublicationInfo': {'members': [{'name': 'action',
                                                                 'type': 'INTEGER'},
                                                                {'element': {'type': 'SinglePubInfo'},
                                                                 'name': 'pubInfos',
                                                                 'optional': True,
                                                                 'size': [(1,
                                                                           'MAX')],
                                                                 'type': 'SEQUENCE '
                                                                         'OF'}],
                                                    'type': 'SEQUENCE'},
                             'PKMACValue': {'members': [{'name': 'algId',
                                                         'type': 'AlgorithmIdentifier'},
                                                        {'name': 'value',
                                                         'type': 'BIT STRING'}],
                                            'type': 'SEQUENCE'},
                             'POPOPrivKey': {'members': [{'name': 'thisMessage',
                                                          'tag': {'number': 0},
                                                          'type': 'BIT STRING'},
                                                         {'name': 'subsequentMessage',
                                                          'tag': {'number': 1},
                                                          'type': 'SubsequentMessage'},
                                                         {'name': 'dhMAC',
                                                          'tag': {'number': 2},
                                                          'type': 'BIT STRING'},
                                                         {'name': 'agreeMAC',
                                                          'tag': {'number': 3},
                                                          'type': 'PKMACValue'},
                                                         {'name': 'encryptedKey',
                                                          'tag': {'number': 4},
                                                          'type': 'EnvelopedData'}],
                                             'type': 'CHOICE'},
                             'POPOSigningKey': {'members': [{'name': 'poposkInput',
                                                             'optional': True,
                                                             'tag': {'number': 0},
                                                             'type': 'POPOSigningKeyInput'},
                                                            {'name': 'algorithmIdentifier',
                                                             'type': 'AlgorithmIdentifier'},
                                                            {'name': 'signature',
                                                             'type': 'BIT '
                                                                     'STRING'}],
                                                'type': 'SEQUENCE'},
                             'POPOSigningKeyInput': {'members': [{'members': [{'name': 'sender',
                                                                               'tag': {'number': 0},
                                                                               'type': 'GeneralName'},
                                                                              {'name': 'publicKeyMAC',
                                                                               'type': 'PKMACValue'}],
                                                                  'name': 'authInfo',
                                                                  'type': 'CHOICE'},
                                                                 {'name': 'publicKey',
                                                                  'type': 'SubjectPublicKeyInfo'}],
                                                     'type': 'SEQUENCE'},
                             'PrivateKeyInfo': {'members': [{'name': 'version',
                                                             'type': 'INTEGER'},
                                                            {'name': 'privateKeyAlgorithm',
                                                             'type': 'AlgorithmIdentifier'},
                                                            {'name': 'privateKey',
                                                             'type': 'OCTET '
                                                                     'STRING'},
                                                            {'name': 'attributes',
                                                             'optional': True,
                                                             'tag': {'kind': 'IMPLICIT',
                                                                     'number': 0},
                                                             'type': 'Attributes'}],
                                                'type': 'SEQUENCE'},
                             'ProofOfPossession': {'members': [{'name': 'raVerified',
                                                                'tag': {'number': 0},
                                                                'type': 'NULL'},
                                                               {'name': 'signature',
                                                                'tag': {'number': 1},
                                                                'type': 'POPOSigningKey'},
                                                               {'name': 'keyEncipherment',
                                                                'tag': {'number': 2},
                                                                'type': 'POPOPrivKey'},
                                                               {'name': 'keyAgreement',
                                                                'tag': {'number': 3},
                                                                'type': 'POPOPrivKey'}],
                                                   'type': 'CHOICE'},
                             'ProtocolEncrKey': {'type': 'SubjectPublicKeyInfo'},
                             'RegToken': {'type': 'UTF8String'},
                             'SinglePubInfo': {'members': [{'name': 'pubMethod',
                                                            'type': 'INTEGER'},
                                                           {'name': 'pubLocation',
                                                            'optional': True,
                                                            'type': 'GeneralName'}],
                                               'type': 'SEQUENCE'},
                             'SubsequentMessage': {'type': 'INTEGER'},
                             'UTF8Pairs': {'type': 'UTF8String'}},
                   'values': {'id-ct': {'type': 'OBJECT IDENTIFIER',
                                        'value': ['id-smime', 1]},
                              'id-ct-encKeyWithID': {'type': 'OBJECT '
                                                             'IDENTIFIER',
                                                     'value': ['id-ct', 21]},
                              'id-pkip': {'type': 'OBJECT IDENTIFIER',
                                          'value': ['id-pkix', 5]},
                              'id-pkix': {'type': 'OBJECT IDENTIFIER',
                                          'value': [('iso', 1),
                                                    ('identified-organization',
                                                     3),
                                                    ('dod', 6),
                                                    ('internet', 1),
                                                    ('security', 5),
                                                    ('mechanisms', 5),
                                                    7]},
                              'id-regCtrl': {'type': 'OBJECT IDENTIFIER',
                                             'value': ['id-pkip', 1]},
                              'id-regCtrl-authenticator': {'type': 'OBJECT '
                                                                   'IDENTIFIER',
                                                           'value': ['id-regCtrl',
                                                                     2]},
                              'id-regCtrl-oldCertID': {'type': 'OBJECT '
                                                               'IDENTIFIER',
                                                       'value': ['id-regCtrl',
                                                                 5]},
                              'id-regCtrl-pkiArchiveOptions': {'type': 'OBJECT '
                                                                       'IDENTIFIER',
                                                               'value': ['id-regCtrl',
                                                                         4]},
                              'id-regCtrl-pkiPublicationInfo': {'type': 'OBJECT '
                                                                        'IDENTIFIER',
                                                                'value': ['id-regCtrl',
                                                                          3]},
                              'id-regCtrl-protocolEncrKey': {'type': 'OBJECT '
                                                                     'IDENTIFIER',
                                                             'value': ['id-regCtrl',
                                                                       6]},
                              'id-regCtrl-regToken': {'type': 'OBJECT '
                                                              'IDENTIFIER',
                                                      'value': ['id-regCtrl',
                                                                1]},
                              'id-regInfo': {'type': 'OBJECT IDENTIFIER',
                                             'value': ['id-pkip', 2]},
                              'id-regInfo-certReq': {'type': 'OBJECT '
                                                             'IDENTIFIER',
                                                     'value': ['id-regInfo',
                                                               2]},
                              'id-regInfo-utf8Pairs': {'type': 'OBJECT '
                                                               'IDENTIFIER',
                                                       'value': ['id-regInfo',
                                                                 1]},
                              'id-smime': {'type': 'OBJECT IDENTIFIER',
                                           'value': [('iso', 1),
                                                     ('member-body', 2),
                                                     ('us', 840),
                                                     ('rsadsi', 113549),
                                                     ('pkcs', 1),
                                                     ('pkcs9', 9),
                                                     16]}}}}


"""Exceptions.

"""


class Error(Exception):
    """Base exception of all asn1tools exceptions.

    """


class CompileError(Error):
    """General ASN.1 compile error.

    """


class EncodeError(Error):
    """General ASN.1 encode error.

    """


class DecodeError(Error):
    """General ASN.1 decode error.

    """


class ConstraintsError(Error):
    """General ASN.1 constraints.

    """



EXPECTED = {'X691-A1': {'extensibility-implied': False,
             'imports': {},
             'object-classes': {},
             'object-sets': {},
             'types': {'ChildInformation': {'members': [{'name': 'name',
                                                         'type': 'Name'},
                                                        {'name': 'dateOfBirth',
                                                         'tag': {'number': 0},
                                                         'type': 'Date'}],
                                            'type': 'SET'},
                       'Date': {'tag': {'class': 'APPLICATION',
                                        'kind': 'IMPLICIT',
                                        'number': 3},
                                'type': 'VisibleString'},
                       'EmployeeNumber': {'tag': {'class': 'APPLICATION',
                                                  'kind': 'IMPLICIT',
                                                  'number': 2},
                                          'type': 'INTEGER'},
                       'Name': {'members': [{'name': 'givenName',
                                             'type': 'VisibleString'},
                                            {'name': 'initial',
                                             'type': 'VisibleString'},
                                            {'name': 'familyName',
                                             'type': 'VisibleString'}],
                                'tag': {'class': 'APPLICATION',
                                        'kind': 'IMPLICIT',
                                        'number': 1},
                                'type': 'SEQUENCE'},
                       'PersonnelRecord': {'members': [{'name': 'name',
                                                        'type': 'Name'},
                                                       {'name': 'title',
                                                        'tag': {'number': 0},
                                                        'type': 'VisibleString'},
                                                       {'name': 'number',
                                                        'type': 'EmployeeNumber'},
                                                       {'name': 'dateOfHire',
                                                        'tag': {'number': 1},
                                                        'type': 'Date'},
                                                       {'name': 'nameOfSpouse',
                                                        'tag': {'number': 2},
                                                        'type': 'Name'},
                                                       {'default': [],
                                                        'element': {'type': 'ChildInformation'},
                                                        'name': 'children',
                                                        'tag': {'kind': 'IMPLICIT',
                                                                'number': 3},
                                                        'type': 'SEQUENCE OF'}],
                                           'tag': {'class': 'APPLICATION',
                                                   'kind': 'IMPLICIT',
                                                   'number': 0},
                                           'type': 'SET'}},
             'values': {}}}


"""Unaligned Packed Encoding Rules (UPER) codec.

"""

import logging
import string

from . import EncodeError
from . import DecodeError
from . import per
from .ber import encode_object_identifier
from .ber import decode_object_identifier
from .per import integer_as_number_of_bits
from .per import PermittedAlphabet
from .per import Encoder
from .per import Type
from .per import Boolean
from .per import Null
from .per import Any
from .per import Enumerated
from .per import Recursive
from .per import Real


LOGGER = logging.getLogger(__name__)


class KnownMultiplierStringType(Type):

    PERMITTED_ALPHABET = ''

    def __init__(self,
                 name,
                 minimum=None,
                 maximum=None,
                 has_extension_marker=None,
                 permitted_alphabet=None):
        super(KnownMultiplierStringType, self).__init__(name,
                                                        self.__class__.__name__)
        self.set_size_range(minimum, maximum, has_extension_marker)

        if permitted_alphabet is None:
            permitted_alphabet = self.PERMITTED_ALPHABET

        self.permitted_alphabet = permitted_alphabet
        self.bits_per_character = integer_as_number_of_bits(
            len(permitted_alphabet) - 1)

    def set_size_range(self, minimum, maximum, has_extension_marker):
        self.minimum = minimum
        self.maximum = maximum
        self.has_extension_marker = has_extension_marker

        if minimum is None or maximum is None:
            self.number_of_bits = None
        else:
            size = maximum - minimum
            self.number_of_bits = integer_as_number_of_bits(size)

    def encode(self, data, encoder):
        encoded = data.encode('ascii')

        if self.has_extension_marker:
            encoder.append_bit(0)

        if self.number_of_bits is None:
            encoder.append_length_determinant(len(encoded))
        elif self.minimum != self.maximum:
            encoder.append_non_negative_binary_integer(len(encoded) - self.minimum,
                                                       self.number_of_bits)

        for value in bytearray(encoded):
            encoder.append_non_negative_binary_integer(
                self.permitted_alphabet.encode(value),
                self.bits_per_character)

    def decode(self, decoder):
        if self.has_extension_marker:
            bit = decoder.read_bit()

            if bit:
                raise NotImplementedError('Extension is not implemented.')

        if self.number_of_bits is None:
            length = decoder.read_length_determinant()
        else:
            length = self.minimum

            if self.minimum != self.maximum:
                length += decoder.read_non_negative_binary_integer(self.number_of_bits)

        data = []

        for _ in range(length):
            value = decoder.read_non_negative_binary_integer(self.bits_per_character)
            data.append(self.permitted_alphabet.decode(value))

        return bytearray(data).decode('ascii')

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               self.name)


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name, 'INTEGER')
        self.minimum = None
        self.maximum = None
        self.has_extension_marker = None
        self.number_of_bits = None

    def set_restricted_to_range(self, minimum, maximum, has_extension_marker):
        self.minimum = minimum
        self.maximum = maximum
        self.has_extension_marker = has_extension_marker

        if minimum is None or maximum is None:
            self.number_of_bits = None
        else:
            size = self.maximum - self.minimum
            self.number_of_bits = integer_as_number_of_bits(size)

    def encode(self, data, encoder):
        if self.has_extension_marker:
            encoder.append_bit(0)

        if self.number_of_bits is None:
            encoder.append_unconstrained_whole_number(data)
        else:
            encoder.append_non_negative_binary_integer(data - self.minimum,
                                                       self.number_of_bits)

    def decode(self, decoder):
        if self.has_extension_marker:
            bit = decoder.read_bit()

            if bit:
                raise NotImplementedError('Extension is not implemented.')

        if self.number_of_bits is None:
            length = decoder.read_length_determinant()
            value = decoder.read_unconstrained_whole_number(length)
        else:
            value = decoder.read_non_negative_binary_integer(self.number_of_bits)
            value += self.minimum

        return value

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class MembersType(Type):

    def __init__(self,
                 name,
                 root_members,
                 additions,
                 type_name):
        super(MembersType, self).__init__(name, type_name)
        self.root_members = root_members
        self.additions = additions
        self.optionals = [
            member
            for member in root_members
            if member.optional or member.default is not None
        ]

    def encode(self, data, encoder):
        if self.additions is not None:
            offset = encoder.number_of_bits
            encoder.append_bit(0)
            self.encode_root(data, encoder)

            if len(self.additions) > 0:
                if self.encode_additions(data, encoder):
                    encoder.set_bit(offset)
        else:
            self.encode_root(data, encoder)

    def encode_root(self, data, encoder):
        for optional in self.optionals:
            if optional.optional:
                encoder.append_bit(optional.name in data)
            elif optional.name in data:
                encoder.append_bit(data[optional.name] != optional.default)
            else:
                encoder.append_bit(0)

        for member in self.root_members:
            self.encode_member(member, data, encoder)

    def encode_additions(self, data, encoder):
        # Encode extension additions.
        presence_bits = 0
        addition_encoders = []

        try:
            for addition in self.additions:
                presence_bits <<= 1
                addition_encoder = Encoder()

                if isinstance(addition, AdditionGroup):
                    addition.encode_addition_group(data, addition_encoder)
                else:
                    self.encode_member(addition,
                                       data,
                                       addition_encoder,
                                       encode_default=True)

                if addition_encoder.number_of_bits > 0:
                    addition_encoders.append(addition_encoder)
                    presence_bits |= 1
        except EncodeError:
            pass

        # Return false if no extension additions are present.
        if not addition_encoders:
            return False

        # Presence bit field.
        number_of_additions = len(self.additions)
        encoder.append_normally_small_length(number_of_additions)
        encoder.append_non_negative_binary_integer(presence_bits,
                                                   number_of_additions)

        # Embed each encoded extension addition in an open type (add a
        # length field and multiple of 8 bits).
        for addition_encoder in addition_encoders:
            addition_encoder.align()
            encoder.append_length_determinant(addition_encoder.number_of_bytes())
            encoder += addition_encoder

        return True

    def encode_addition_group(self, data, encoder):
        self.encode_root(data, encoder)

        if ((encoder.value == 0)
            and (encoder.number_of_bits == len(self.optionals))):
            encoder.number_of_bits = 0

    def encode_member(self, member, data, encoder, encode_default=False):
        name = member.name

        if name in data:
            if member.default is None:
                member.encode(data[name], encoder)
            elif data[name] != member.default or encode_default:
                member.encode(data[name], encoder)
        elif member.optional or member.default is not None:
            pass
        else:
            raise EncodeError(
                "{} member '{}' not found in {}.".format(
                    self.__class__.__name__,
                    name,
                    data))

    def decode(self, decoder):
        if self.additions is not None:
            if decoder.read_bit():
                decoded = self.decode_root(decoder)
                decoded.update(self.decode_additions(decoder))

                return decoded
            else:
                return self.decode_root(decoder)
        else:
            return self.decode_root(decoder)

    def decode_root(self, decoder):
        values = {}
        optionals = {
            optional: decoder.read_bit()
            for optional in self.optionals
        }

        for member in self.root_members:
            try:
                if optionals.get(member, True):
                    value = member.decode(decoder)
                    values[member.name] = value
                elif member.default is not None:
                    values[member.name] = member.default
            except DecodeError as e:
                e.location.append(member.name)
                raise

        return values

    def decode_additions(self, decoder):
        # Presence bit field.
        length = decoder.read_normally_small_length()
        presence_bits = decoder.read_non_negative_binary_integer(length)
        decoded = {}

        for i in range(length):
            if presence_bits & (1 << (length - i - 1)):
                # Open type decoding.
                open_type_length = decoder.read_length_determinant()
                offset = decoder.number_of_bits

                if i < len(self.additions):
                    addition = self.additions[i]

                    if isinstance(addition, AdditionGroup):
                        decoded.update(addition.decode(decoder))
                    else:
                        try:
                            decoded[addition.name] = addition.decode(decoder)
                        except DecodeError as e:
                            e.location.append(addition.name)
                            raise
                else:
                    decoder.skip_bits(8 * open_type_length)

                alignment_bits = (offset - decoder.number_of_bits) % 8

                if alignment_bits != 0:
                    decoder.skip_bits(8 - alignment_bits)

        return decoded

    def __repr__(self):
        return '{}({}, [{}])'.format(
            self.__class__.__name__,
            self.name,
            ', '.join([repr(member) for member in self.root_members]))


class Sequence(MembersType):

    def __init__(self,
                 name,
                 root_members,
                 additions):
        super(Sequence, self).__init__(name,
                                       root_members,
                                       additions,
                                       'SEQUENCE')


class Set(MembersType):

    def __init__(self,
                 name,
                 root_members,
                 additions):
        super(Set, self).__init__(name,
                                  root_members,
                                  additions,
                                  'SET')


class AdditionGroup(Sequence):
    pass


class ArrayType(Type):

    def __init__(self,
                 name,
                 element_type,
                 minimum,
                 maximum,
                 has_extension_marker,
                 type_name):
        super(ArrayType, self).__init__(name, type_name)
        self.element_type = element_type
        self.minimum = minimum
        self.maximum = maximum
        self.has_extension_marker = has_extension_marker

        if minimum is None or maximum is None:
            self.number_of_bits = None
        else:
            size = maximum - minimum
            self.number_of_bits = integer_as_number_of_bits(size)

    def encode(self, data, encoder):
        if self.has_extension_marker:
            encoder.append_bit(0)

        if self.number_of_bits is None:
            encoder.append_length_determinant(len(data))
        elif self.minimum != self.maximum:
            encoder.append_non_negative_binary_integer(
                len(data) - self.minimum,
                self.number_of_bits)

        for entry in data:
            self.element_type.encode(entry, encoder)

    def decode(self, decoder):
        if self.has_extension_marker:
            bit = decoder.read_bit()

            if bit:
                raise NotImplementedError('Extension is not implemented.')

        if self.number_of_bits is None:
            length = decoder.read_length_determinant()
        else:
            length = self.minimum

            if self.minimum != self.maximum:
                length += decoder.read_non_negative_binary_integer(self.number_of_bits)

        decoded = []

        for _ in range(length):
            decoded_element = self.element_type.decode(decoder)
            decoded.append(decoded_element)

        return decoded

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.name,
                                   self.element_type)


class SequenceOf(ArrayType):

    def __init__(self,
                 name,
                 element_type,
                 minimum,
                 maximum,
                 has_extension_marker):
        super(SequenceOf, self).__init__(name,
                                         element_type,
                                         minimum,
                                         maximum,
                                         has_extension_marker,
                                         'SEQUENCE OF')


class SetOf(ArrayType):

    def __init__(self,
                 name,
                 element_type,
                 minimum,
                 maximum,
                 has_extension_marker):
        super(SetOf, self).__init__(name,
                                    element_type,
                                    minimum,
                                    maximum,
                                    has_extension_marker,
                                    'SET OF')


class BitString(Type):

    def __init__(self, name, minimum, maximum, has_named_bits):
        super(BitString, self).__init__(name, 'BIT STRING')
        self.minimum = minimum
        self.maximum = maximum
        self.has_named_bits = has_named_bits

        if minimum is None or maximum is None:
            self.number_of_bits = None
        else:
            size = self.maximum - self.minimum
            self.number_of_bits = integer_as_number_of_bits(size)

    def rstrip_zeros(self, data, number_of_bits):
        data = bytearray(data.rstrip(b'\x00'))

        if len(data) == 0:
            number_of_bits = 0
        else:
            value = data[-1]
            number_of_bits = 8 * len(data) - (value & -value).bit_length() + 1

        if self.minimum is not None:
            if number_of_bits < self.minimum:
                number_of_bits = self.minimum
                number_of_bytes = ((number_of_bits + 7) // 8)
                data += (number_of_bytes - len(data)) * b'\x00'

        return data, number_of_bits

    def encode(self, data, encoder):
        data, number_of_bits = data

        if self.has_named_bits:
            data, number_of_bits = self.rstrip_zeros(data, number_of_bits)

        if self.number_of_bits is None:
            encoder.append_length_determinant(number_of_bits)
        elif self.minimum != self.maximum:
            encoder.append_non_negative_binary_integer(
                number_of_bits - self.minimum,
                self.number_of_bits)

        encoder.append_bits(data, number_of_bits)

    def decode(self, decoder):
        if self.number_of_bits is None:
            number_of_bits = decoder.read_length_determinant()
        else:
            number_of_bits = self.minimum

            if self.minimum != self.maximum:
                number_of_bits += decoder.read_non_negative_binary_integer(
                    self.number_of_bits)

        value = decoder.read_bits(number_of_bits)

        return (value, number_of_bits)

    def __repr__(self):
        return 'BitString({})'.format(self.name)


class OctetString(Type):

    def __init__(self, name, minimum, maximum):
        super(OctetString, self).__init__(name, 'OCTET STRING')
        self.minimum = minimum
        self.maximum = maximum

        if minimum is None or maximum is None:
            self.number_of_bits = None
        else:
            size = self.maximum - self.minimum
            self.number_of_bits = integer_as_number_of_bits(size)

    def encode(self, data, encoder):
        if self.number_of_bits is None:
            encoder.append_length_determinant(len(data))
        elif self.minimum != self.maximum:
            encoder.append_non_negative_binary_integer(len(data) - self.minimum,
                                                       self.number_of_bits)

        encoder.append_bytes(data)

    def decode(self, decoder):
        if self.number_of_bits is None:
            length = decoder.read_length_determinant()
        else:
            length = self.minimum

            if self.minimum != self.maximum:
                length += decoder.read_non_negative_binary_integer(
                    self.number_of_bits)

        return decoder.read_bits(8 * length)

    def __repr__(self):
        return 'OctetString({})'.format(self.name)


class IA5String(KnownMultiplierStringType):

    ENCODE_DECODE_MAP = {v: v for v in range(128)}
    PERMITTED_ALPHABET = PermittedAlphabet(ENCODE_DECODE_MAP,
                                           ENCODE_DECODE_MAP)


class NumericString(KnownMultiplierStringType):

    ALPHABET = bytearray(b' 0123456789')
    ENCODE_MAP = {v: i for i, v in enumerate(ALPHABET)}
    DECODE_MAP = {i: v for i, v in enumerate(ALPHABET)}
    PERMITTED_ALPHABET = PermittedAlphabet(ENCODE_MAP,
                                           DECODE_MAP)


class PrintableString(KnownMultiplierStringType):

    ALPHABET = bytearray((string.ascii_uppercase
                          + string.ascii_lowercase
                          + string.digits
                          + " '()+,-./:=?").encode('ascii'))
    ENCODE_MAP = {v: v for v in ALPHABET}
    DECODE_MAP = {v: v for v in ALPHABET}
    PERMITTED_ALPHABET = PermittedAlphabet(ENCODE_MAP,
                                           DECODE_MAP)


class UniversalString(Type):

    def __init__(self, name):
        super(UniversalString, self).__init__(name, 'UniversalString')

    def encode(self, _data, _encoder):
        raise NotImplementedError('UniversalString not yet implemented.')

    def decode(self, _decoder):
        raise NotImplementedError('UniversalString not yet implemented.')

    def __repr__(self):
        return 'UniversalString({})'.format(self.name)


class VisibleString(KnownMultiplierStringType):

    ENCODE_DECODE_MAP = {v: v for v in range(32, 127)}
    PERMITTED_ALPHABET = PermittedAlphabet(ENCODE_DECODE_MAP,
                                           ENCODE_DECODE_MAP)


class GeneralString(Type):

    def __init__(self, name):
        super(GeneralString, self).__init__(name, 'GeneralString')

    def encode(self, _data, _encoder):
        raise NotImplementedError('GeneralString not yet implemented.')

    def decode(self, _decoder):
        raise NotImplementedError('GeneralString not yet implemented.')

    def __repr__(self):
        return 'GeneralString({})'.format(self.name)


class UTF8String(Type):

    def __init__(self, name):
        super(UTF8String, self).__init__(name, 'UTF8String')

    def encode(self, data, encoder):
        encoded = data.encode('utf-8')
        encoder.append_length_determinant(len(encoded))
        encoder.append_bytes(bytearray(encoded))

    def decode(self, decoder):
        length = decoder.read_length_determinant()
        encoded = decoder.read_bits(8 * length)

        return encoded.decode('utf-8')

    def __repr__(self):
        return 'UTF8String({})'.format(self.name)


class GraphicString(Type):

    def __init__(self, name):
        super(GraphicString, self).__init__(name, 'GraphicString')

    def encode(self, data, encoder):
        encoded = data.encode('latin-1')
        encoder.append_length_determinant(len(encoded))
        encoder.append_bytes(bytearray(encoded))

    def decode(self, decoder):
        length = decoder.read_length_determinant()
        encoded = decoder.read_bits(8 * length)

        return encoded.decode('latin-1')

    def __repr__(self):
        return 'GraphicString({})'.format(self.name)


class BMPString(Type):

    def __init__(self, name):
        super(BMPString, self).__init__(name, 'BMPString')

    def encode(self, _data, _encoder):
        raise NotImplementedError('BMPString not yet implemented.')

    def decode(self, _decoder):
        raise NotImplementedError('BMPString not yet implemented.')

    def __repr__(self):
        return 'BMPString({})'.format(self.name)


class TeletexString(Type):

    def __init__(self, name):
        super(TeletexString, self).__init__(name, 'TeletexString')

    def encode(self, _data, _encoder):
        raise NotImplementedError('TeletexString not yet implemented.')

    def decode(self, _decoder):
        raise NotImplementedError('TeletexString not yet implemented.')

    def __repr__(self):
        return 'TeletexString({})'.format(self.name)


class UTCTime(VisibleString):
    pass


class GeneralizedTime(VisibleString):
    pass


class ObjectIdentifier(per.ObjectIdentifier):

    def encode(self, data, encoder):
        encoded_subidentifiers = encode_object_identifier(data)
        encoder.append_length_determinant(len(encoded_subidentifiers))
        encoder.append_bytes(bytearray(encoded_subidentifiers))

    def decode(self, decoder):
        length = decoder.read_length_determinant()
        data = decoder.read_bits(8 * length)

        return decode_object_identifier(bytearray(data), 0, len(data))


class Choice(per.Choice):

    def encode_additions(self, data, encoder):
        try:
            index = self.additions_name_to_index[data[0]]
        except KeyError:
            raise EncodeError(
                "Expected choices are {}, but got '{}'.".format(
                    sorted([member.name for member in self.all_members()]),
                    data[0]))

        addition_encoder = Encoder()
        addition = self.additions_index_to_member[index]
        addition.encode(data[1], addition_encoder)

        # Embed encoded extension addition in an open type (add a
        # length field and multiple of 8 bits).
        addition_encoder.align()
        encoder.append_normally_small_non_negative_whole_number(index)
        encoder.append_length_determinant(addition_encoder.number_of_bytes())
        encoder += addition_encoder

    def decode_additions(self, decoder):
        index = decoder.read_normally_small_non_negative_whole_number()
        addition = self.additions_index_to_member[index]

        # Open type decoding.
        decoder.read_length_determinant()
        offset = decoder.number_of_bits
        decoded = addition.decode(decoder)
        alignment_bits = (offset - decoder.number_of_bits) % 8

        if alignment_bits != 0:
            decoder.skip_bits(8 - alignment_bits)

        return (addition.name, decoded)


class Compiler(per.Compiler):

    def compile_type(self, name, type_descriptor, module_name):
        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            compiled = Sequence(name,
                                *self.compile_members(type_descriptor['members'],
                                                      module_name))
        elif type_name == 'SEQUENCE OF':
            compiled = SequenceOf(name,
                                  self.compile_type('',
                                                    type_descriptor['element'],
                                                    module_name),
                                  *self.get_size_range(type_descriptor,
                                                       module_name))
        elif type_name == 'SET':
            compiled = Set(name,
                           *self.compile_members(
                               type_descriptor['members'],
                               module_name,
                               sort_by_tag=True))
        elif type_name == 'SET OF':
            compiled = SetOf(name,
                             self.compile_type('',
                                               type_descriptor['element'],
                                               module_name),
                             *self.get_size_range(type_descriptor,
                                                  module_name))
        elif type_name == 'CHOICE':
            compiled = Choice(name,
                              *self.compile_members(
                                  type_descriptor['members'],
                                  module_name,
                                  flat_additions=True))
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name, type_descriptor['values'])
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            minimum, maximum, _ = self.get_size_range(type_descriptor,
                                                      module_name)
            compiled = OctetString(name, minimum, maximum)
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            compiled = NumericString(name,
                                     *self.get_size_range(type_descriptor,
                                                          module_name))
        elif type_name == 'PrintableString':
            compiled = PrintableString(name,
                                       *self.get_size_range(type_descriptor,
                                                            module_name))
        elif type_name == 'IA5String':
            compiled = IA5String(name,
                                 *self.get_size_range(type_descriptor,
                                                      module_name))
        elif type_name == 'VisibleString':
            permitted_alphabet = self.get_permitted_alphabet(type_descriptor)
            compiled = VisibleString(name,
                                     *self.get_size_range(type_descriptor,
                                                          module_name),
                                     permitted_alphabet=permitted_alphabet)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'GraphicString':
            compiled = GraphicString(name)
        elif type_name == 'BMPString':
            compiled = BMPString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'BIT STRING':
            minimum, maximum, _ = self.get_size_range(type_descriptor,
                                                      module_name)
            has_named_bits = ('named-bits' in type_descriptor)
            compiled = BitString(name,
                                 minimum,
                                 maximum,
                                 has_named_bits)
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            compiled = Any(name)
        elif type_name == 'NULL':
            compiled = Null(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
                self.recursive_types.append(compiled)
            else:
                compiled = self.compile_user_type(name,
                                                  type_name,
                                                  module_name)

        if 'tag' in type_descriptor:
            compiled = self.set_compiled_tag(compiled, type_descriptor)

        if 'restricted-to' in type_descriptor:
            compiled = self.set_compiled_restricted_to(compiled,
                                                       type_descriptor,
                                                       module_name)

        return compiled

    def compile_extension_member(self,
                                 member,
                                 module_name,
                                 additions,
                                 flat_additions):
        if isinstance(member, list):
            if flat_additions:
                for memb in member:
                    compiled_member = self.compile_member(memb,
                                                          module_name)
                    additions.append(compiled_member)
            else:
                compiled_member, _ = self.compile_members(member,
                                                          module_name)
                compiled_group = AdditionGroup('ExtensionAddition',
                                               compiled_member,
                                               None)
                additions.append(compiled_group)
        else:
            compiled_member = self.compile_member(member,
                                                  module_name)
            additions.append(compiled_member)


def compile_dict(specification):
    return Compiler(specification).process()


def decode_length(_data):
    raise DecodeError('Decode length is not supported for this codec.')



"""JSON Encoding Rules (JER) codec.

"""

import json
import binascii
import math
import logging

from . import EncodeError
from . import DecodeError
from . import compiler
from .compiler import enum_values_as_dict


LOGGER = logging.getLogger(__name__)


class Type(object):

    def __init__(self, name, type_name):
        self.name = name
        self.type_name = type_name
        self.optional = False
        self.default = None

    def set_size_range(self, minimum, maximum, has_extension_marker):
        pass


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name, 'INTEGER')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class Real(Type):

    def __init__(self, name):
        super(Real, self).__init__(name, 'REAL')

    def encode(self, data):
        if data == float('inf'):
            return 'INF'
        elif data == float('-inf'):
            return '-INF'
        elif math.isnan(data):
            return 'NaN'
        else:
            return data

    def decode(self, data):
        if isinstance(data, float):
            return data
        else:
            return {
                'INF': float('inf'),
                '-INF': float('-inf'),
                'NaN': float('nan'),
                '0': 0.0,
                '-0': 0.0
            }[data]

    def __repr__(self):
        return 'Real({})'.format(self.name)


class Boolean(Type):

    def __init__(self, name):
        super(Boolean, self).__init__(name, 'BOOLEAN')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'Boolean({})'.format(self.name)


class IA5String(Type):

    def __init__(self, name):
        super(IA5String, self).__init__(name, 'IA5String')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'IA5String({})'.format(self.name)


class NumericString(Type):

    def __init__(self, name):
        super(NumericString, self).__init__(name, 'NumericString')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'NumericString({})'.format(self.name)


class Sequence(Type):

    def __init__(self, name, members):
        super(Sequence, self).__init__(name, 'SEQUENCE')
        self.members = members

    def encode(self, data):
        values = {}

        for member in self.members:
            name = member.name

            if name in data:
                value = member.encode(data[name])
            elif member.optional or member.default is not None:
                continue
            else:
                raise EncodeError(
                    "Sequence member '{}' not found in {}.".format(
                        name,
                        data))

            values[name] = value

        return values

    def decode(self, data):
        values = {}

        for member in self.members:
            name = member.name

            if name in data:
                value = member.decode(data[name])
                values[name] = value
            elif member.optional:
                pass
            elif member.default is not None:
                values[name] = member.default

        return values

    def __repr__(self):
        return 'Sequence({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Set(Type):

    def __init__(self, name, members):
        super(Set, self).__init__(name, 'SET')
        self.members = members

    def encode(self, data):
        values = {}

        for member in self.members:
            name = member.name

            if name in data:
                if member.default is None:
                    value = member.encode(data[name])
                elif data[name] != member.default:
                    value = member.encode(data[name])
            elif member.optional or member.default is not None:
                continue
            else:
                raise EncodeError(
                    "Set member '{}' not found in {}.".format(
                        name,
                        data))

            values[name] = value

        return values

    def decode(self, data):
        values = {}

        for member in self.members:
            name = member.name

            if name in data:
                value = member.decode(data[name])
                values[name] = value
            elif member.optional:
                pass
            elif member.default is not None:
                values[name] = member.default

        return values

    def __repr__(self):
        return 'Set({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class SequenceOf(Type):

    def __init__(self, name, element_type):
        super(SequenceOf, self).__init__(name, 'SEQUENCE OF')
        self.element_type = element_type

    def encode(self, data):
        values = []

        for entry in data:
            value = self.element_type.encode(entry)
            values.append(value)

        return values

    def decode(self, data):
        values = []

        for element_data in data:
            value = self.element_type.decode(element_data)
            values.append(value)

        return values

    def __repr__(self):
        return 'SequenceOf({}, {})'.format(self.name,
                                           self.element_type)


class SetOf(Type):

    def __init__(self, name, element_type):
        super(SetOf, self).__init__(name, 'SET OF')
        self.element_type = element_type

    def encode(self, data):
        values = []

        for entry in data:
            value = self.element_type.encode(entry)
            values.append(value)

        return values

    def decode(self, data):
        values = []

        for element_data in data:
            value = self.element_type.decode(element_data)
            values.append(value)

        return values

    def __repr__(self):
        return 'SetOf({}, {})'.format(self.name,
                                      self.element_type)


class BitString(Type):

    def __init__(self, name, minimum, maximum):
        super(BitString, self).__init__(name, 'BIT STRING')

        if minimum is None and maximum is None:
            self.size = None
        elif minimum == maximum:
            self.size = minimum
        else:
            self.size = None

    def encode(self, data):
        value = binascii.hexlify(data[0]).decode('ascii')

        if self.size is None:
            value = {
                "value": value,
                "length": data[1]
            }

        return value

    def decode(self, data):
        if self.size is None:
            return (binascii.unhexlify(data['value']), data['length'])
        else:
            return (binascii.unhexlify(data), self.size)

    def __repr__(self):
        return 'BitString({})'.format(self.name)


class OctetString(Type):

    def __init__(self, name):
        super(OctetString, self).__init__(name, 'OCTET STRING')

    def encode(self, data):
        return binascii.hexlify(data).decode('ascii')

    def decode(self, data):
        return binascii.unhexlify(data)

    def __repr__(self):
        return 'OctetString({})'.format(self.name)


class PrintableString(Type):

    def __init__(self, name):
        super(PrintableString, self).__init__(name, 'PrintableString')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'PrintableString({})'.format(self.name)


class UniversalString(Type):

    def __init__(self, name):
        super(UniversalString, self).__init__(name, 'UniversalString')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'UniversalString({})'.format(self.name)


class VisibleString(Type):

    def __init__(self, name):
        super(VisibleString, self).__init__(name, 'VisibleString')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'VisibleString({})'.format(self.name)


class GeneralString(Type):

    def __init__(self, name):
        super(GeneralString, self).__init__(name, 'GeneralString')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'GeneralString({})'.format(self.name)


class UTF8String(Type):

    def __init__(self, name):
        super(UTF8String, self).__init__(name, 'UTF8String')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'UTF8String({})'.format(self.name)


class BMPString(Type):

    def __init__(self, name):
        super(BMPString, self).__init__(name, 'BMPString')

    def encode(self, data):
        return data.decode('ascii')

    def decode(self, data):
        return data.encode('ascii')

    def __repr__(self):
        return 'BMPString({})'.format(self.name)


class GraphicString(Type):

    def __init__(self, name):
        super(GraphicString, self).__init__(name, 'GraphicString')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'GraphicString({})'.format(self.name)


class UTCTime(Type):

    def __init__(self, name):
        super(UTCTime, self).__init__(name, 'UTCTime')

    def encode(self, data):
        raise NotImplementedError('UTCTime is not yet implemented.')

    def decode(self, data):
        raise NotImplementedError('UTCTime is not yet implemented.')

    def __repr__(self):
        return 'UTCTime({})'.format(self.name)


class GeneralizedTime(Type):

    def __init__(self, name):
        super(GeneralizedTime, self).__init__(name, 'GeneralizedTime')

    def encode(self, data):
        raise NotImplementedError('GeneralizedTime is not yet implemented.')

    def decode(self, data):
        raise NotImplementedError('GeneralizedTime is not yet implemented.')

    def __repr__(self):
        return 'GeneralizedTime({})'.format(self.name)


class TeletexString(Type):

    def __init__(self, name):
        super(TeletexString, self).__init__(name, 'TeletexString')

    def encode(self, data):
        return data.decode('ascii')

    def decode(self, data):
        return data.encode('ascii')

    def __repr__(self):
        return 'TeletexString({})'.format(self.name)


class ObjectIdentifier(Type):

    def __init__(self, name):
        super(ObjectIdentifier, self).__init__(name, 'OBJECT IDENTIFIER')

    def encode(self, data):
        raise NotImplementedError('OBJECT IDENTIFIER is not yet implemented.')

    def decode(self, data):
        raise NotImplementedError('OBJECT IDENTIFIER is not yet implemented.')

    def __repr__(self):
        return 'ObjectIdentifier({})'.format(self.name)


class Choice(Type):

    def __init__(self, name, members):
        super(Choice, self).__init__(name, 'CHOICE')
        self.members = members
        self.name_to_member = {member.name: member for member in self.members}

    def encode(self, data):
        try:
            member = self.name_to_member[data[0]]
        except KeyError:
            raise EncodeError(
                "Expected choices are {}, but got '{}'.".format(
                    sorted([member.name for member in self.members]),
                    data[0]))

        return {member.name: member.encode(data[1])}

    def decode(self, data):
        name, value = list(data.items())[0]

        try:
            member = self.name_to_member[name]
        except KeyError:
            raise DecodeError(
                "Expected choices are {}, but got '{}'.".format(
                    sorted([member.name for member in self.members]),
                    name))

        return (name, member.decode(value))

    def __repr__(self):
        return 'Choice({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Null(Type):

    def __init__(self, name):
        super(Null, self).__init__(name, 'NULL')

    def encode(self, data):
        return data

    def decode(self, data):
        return data

    def __repr__(self):
        return 'Null({})'.format(self.name)


class Any(Type):

    def __init__(self, name):
        super(Any, self).__init__(name, 'ANY')

    def encode(self, _, encoder):
        raise NotImplementedError('ANY is not yet implemented.')

    def decode(self, data):
        raise NotImplementedError('ANY is not yet implemented.')

    def __repr__(self):
        return 'Any({})'.format(self.name)


class Enumerated(Type):

    def __init__(self, name, values):
        super(Enumerated, self).__init__(name, 'ENUMERATED')
        self.values = set(enum_values_as_dict(values).values())

    def encode(self, data):
        if data not in self.values:
            raise EncodeError(
                "Enumeration value '{}' not found in {}.".format(
                    data[0],
                    sorted(list(self.values))))

        return data

    def decode(self, data):
        if data not in self.values:
            raise DecodeError(
                "Enumeration value '{}' not found in {}.".format(
                    data,
                    sorted(list(self.values))))

        return data

    def __repr__(self):
        return 'Enumerated({})'.format(self.name)


class Recursive(Type, compiler.Recursive):

    def __init__(self, name, type_name, module_name):
        super(Recursive, self).__init__(name, 'RECURSIVE')
        self.type_name = type_name
        self.module_name = module_name
        self._inner = None

    def set_inner_type(self, inner):
        self._inner = inner

    def encode(self, data):
        return self._inner.encode(data)

    def decode(self, data):
        return self._inner.decode(data)

    def __repr__(self):
        return 'Recursive({})'.format(self.name)


class CompiledType(compiler.CompiledType):

    def __init__(self, type_, constraints):
        super(CompiledType, self).__init__(constraints)
        self._type = type_

    @property
    def type(self):
        return self._type

    def encode(self, data, indent=None):
        dictionary = self._type.encode(data)

        if indent is None:
            string = json.dumps(dictionary, separators=(',', ':'))
        else:
            string = json.dumps(dictionary, indent=indent)

        return string.encode('utf-8')

    def decode(self, data):
        return self._type.decode(json.loads(data.decode('utf-8')))

    def __repr__(self):
        return repr(self._type)


class Compiler(compiler.Compiler):

    def process_type(self, type_name, type_descriptor, module_name):
        compiled_type = self.compile_type(type_name,
                                          type_descriptor,
                                          module_name)
        constraints = self.compile_constraints(type_name,
                                               type_descriptor,
                                               module_name)

        return CompiledType(compiled_type, constraints)

    def compile_type(self, name, type_descriptor, module_name):
        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            members = self.compile_members(
                type_descriptor['members'],
                module_name)
            compiled = Sequence(name, members)
        elif type_name == 'SEQUENCE OF':
            compiled = SequenceOf(name,
                                  self.compile_type('',
                                                    type_descriptor['element'],
                                                    module_name))
        elif type_name == 'SET':
            members = self.compile_members(
                type_descriptor['members'],
                module_name)
            compiled = Set(name, members)
        elif type_name == 'SET OF':
            compiled = SetOf(name,
                             self.compile_type('',
                                               type_descriptor['element'],
                                               module_name))
        elif type_name == 'CHOICE':
            members = self.compile_members(
                type_descriptor['members'],
                module_name)
            compiled = Choice(name, members)
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name, type_descriptor['values'])
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            compiled = OctetString(name)
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            compiled = NumericString(name)
        elif type_name == 'PrintableString':
            compiled = PrintableString(name)
        elif type_name == 'IA5String':
            compiled = IA5String(name)
        elif type_name == 'VisibleString':
            compiled = VisibleString(name)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'BMPString':
            compiled = BMPString(name)
        elif type_name == 'GraphicString':
            compiled = GraphicString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'BIT STRING':
            minimum, maximum, _ = self.get_size_range(type_descriptor,
                                                      module_name)
            compiled = BitString(name, minimum, maximum)
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            compiled = Any(name)
        elif type_name == 'NULL':
            compiled = Null(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
                self.recursive_types.append(compiled)
            else:
                compiled = self.compile_user_type(name,
                                                  type_name,
                                                  module_name)

        return compiled


def compile_dict(specification):
    return Compiler(specification).process()


def decode_length(_data):
    raise DecodeError('Decode length is not supported for this codec.')
